import pulumi

from pulumi_aws import ec2
from pulumi.output import Output

from typing import Optional, Mapping
from pydantic.dataclasses import dataclass, ConfigDict

from theia import ResourceOptions
from theia.definitions import THEIA_VERSION
from theia._registry import StackRegistry, GlobalTags


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class SubnetSettings:
    cidr_block: Optional[str | Output[str]] = None
    ipv6_cidr_block: Optional[str | Output[str]] = None
    availability_zone: Optional[str] = None
    map_public_ip_on_launch: Optional[bool] = False
    assign_ipv6_address_on_creation: Optional[bool] = False


class Subnet:
    def __new__(
        cls,
        resource_name: str,
        vpc_id: str | Output[str],
        tags: Optional[Mapping[str, str]] = None,
        settings: Optional[SubnetSettings] = SubnetSettings(),
        opts: Optional[ResourceOptions] = None,
    ) -> ec2.Subnet:
        name = {"Name": resource_name}
        merged_tags = THEIA_VERSION | GlobalTags.get() | name
        if tags:
            merged_tags = tags | merged_tags

        subnet_args = ec2.SubnetArgs(
            vpc_id=vpc_id,
            cidr_block=settings.cidr_block,
            ipv6_cidr_block=settings.ipv6_cidr_block,
            availability_zone=settings.availability_zone,
            map_public_ip_on_launch=settings.map_public_ip_on_launch,
            assign_ipv6_address_on_creation=settings.assign_ipv6_address_on_creation,
            tags=merged_tags,
        )

        subnet = ec2.Subnet(resource_name, args=subnet_args, opts=opts)

        export_values = {"id": subnet.id, "arn": subnet.arn}
        if settings.assign_ipv6_address_on_creation:
            export_values[
                "ipv6_cidr_block_association_id"
            ] = subnet.ipv6_cidr_block_association_id

        pulumi.export(f"subnet---{resource_name}", export_values)

        return subnet

    @staticmethod
    def import_(stack: str, resource_name: str) -> ec2.Subnet:
        stack_ref = StackRegistry.get(stack_name=stack)
        subnet = stack_ref.get_output(name=f"subnet---{resource_name}")
        return ec2.Subnet.get(resource_name=resource_name, id=subnet["id"])
